import torch
import torch.nn as nn
import torch.nn.functional as F

def compute_size(input_size, conv_layer):
    kernel_size = conv_layer.kernel_size[0]
    dilation = conv_layer.dilation[0]
    stride = conv_layer.stride[0]
    padding = conv_layer.padding[0]
    new_size = int((input_size + 2*padding - dilation*(kernel_size - 1) - 1)//stride + 1)
    return new_size

class ResNetBlock(nn.Module):
    def __init__(self, input_size, in_channels, filter, kernel):
        super().__init__()
        self.__type__ = 'cnn'
        
        self.conv1 = nn.Conv1d(in_channels = in_channels, out_channels = filter, kernel_size = kernel, padding= kernel//2)
        out_size = compute_size(input_size, self.conv1)
        self.conv2 = nn.Conv1d(in_channels = filter, out_channels = filter, kernel_size = kernel, padding= kernel//2)
        out_size = compute_size(out_size, self.conv2)
        self.conv3 = nn.Conv1d(in_channels = filter, out_channels = filter, kernel_size = kernel, padding= kernel//2)
        out_size = compute_size(out_size, self.conv3)
        # self.norm = nn.BatchNorm1d(filter)
        
        self.norm = nn.LayerNorm([filter, input_size]) #=> More stable
        self.actv = nn.ReLU()
        self.init_weights = False
        self.out_size = out_size
    def forward(self,x):
        residual = self.conv1(x)
        x = self.actv(residual)
        x = self.actv(self.conv2(x))
        x = self.actv(self.conv3(x))
        x = x + residual
        x = self.norm(x)
        return x
    
class ResNet(nn.Module):
    def __init__(self, input_size, n_class = 7):
        torch.manual_seed(28)
        super().__init__()
        self.__type__ = 'cnn'
        self.__arch__ = 'ResNet'
        self.__version__ = 'baseline'

        self.conv = nn.Conv1d(1, 64, 15)
        self.pool0 = nn.MaxPool1d(kernel_size = 2)
        out_size = compute_size(input_size, self.conv) // 2
        self.block1 = ResNetBlock(out_size, 64, 64, 5)
        self.pool1 = nn.MaxPool1d(kernel_size = 2)
        out_size = out_size // 2
        self.block2 = ResNetBlock(252, 64, 128, 5)
        self.pool2 = nn.MaxPool1d(kernel_size = 2)
        out_size = out_size // 2
        self.block3 = ResNetBlock(126, 128, 128, 5)
        self.final_pool = nn.MaxPool1d(kernel_size = 2)
        out_size = out_size // 2
        self.fc = nn.Linear(128*out_size, n_class)
        self.dropout = nn.Dropout(0.3)
        self.init_weights = True
        self.init_weight()

    def init_weight(self):
        blocks = [self.block1, self.block2, self.block3]
        for i in range(3):
            for layer in blocks[i].children():
                if isinstance(layer, nn.Linear) or isinstance(layer, nn.Conv1d):
                    layer.weight = nn.init.kaiming_normal_(layer.weight, nonlinearity= 'relu')
                    layer.bias = nn.init.zeros_(layer.bias)
                elif isinstance(layer, nn.BatchNorm1d):
                    layer.weight.data.fill_(1.0)
                    layer.bias.data.fill_(0.0)
        self.fc.weight = nn.init.kaiming_normal_(self.fc.weight, nonlinearity= 'relu')

    def forward(self,x):
        # First conv layer
        x = self.pool0(self.conv(x))

        # Three conv operations
        x = self.pool1(self.block1(x))
        x = self.pool2(self.block2(x))
        x = self.block3(x)
        x = self.final_pool(x)
        
        # Flatten and output
        x = x.flatten(1)
        x = self.fc(self.dropout(x))
        x = F.log_softmax(x,-1)
        return x 



import sys
from setuptools import setup

setup(
    name='wavelets_pytorch',
    version='0.1',
    author='Tom Runia',
    author_email='tomrunia@gmail.com',
    url='https://github.com/tomrunia/PyTorchWavelets',
    description='Wavelet Transform in PyTorch',
    long_description='Fast CPU/CUDA implementation of the Continuous Wavelet Transform in PyTorch.',
    license='MIT',
    packages=['wavelets_pytorch'],
    scripts=[]
)


# MIT License
#
# Copyright (c) 2018 Tom Runia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-15

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from abc import ABCMeta, abstractmethod

import numpy as np
import scipy.signal
import scipy.optimize

import torch
from torch.autograd import Variable

from .wavelets import Morlet
from .network import TorchFilterBank

##########################################################################################

@six.add_metaclass(ABCMeta)
class WaveletTransformBase(object):
    """

    Base class for the Continuous Wavelet Transform as described in:
        "Torrence & Combo, A Practical Guide to Wavelet Analysis (BAMS, 1998)"

    This class is a abstract super class for child classes:
        WaveletTransform      => implements CWT in SciPy
        WaveletTransformTorch => implements CWT in PyTorch

    For a more detailed explanation of the parameters, the original code serves as reference:
        https://github.com/aaren/wavelets/blob/master/wavelets/transform.py#L145

    """

    def __init__(self, dt=1.0, dj=0.125, wavelet=Morlet(), unbias=False):
        """
        :param dt: float, sample spacing
        :param dj: float, scale distribution parameter
        :param wavelet: wavelet object, see 'wavelets.py'
        :param unbias: boolean, whether to unbias the power spectrum
        """
        self._dt = dt
        self._dj = dj
        self._wavelet = wavelet
        self._unbias = unbias
        self._scale_minimum = self.compute_minimum_scale()
        self._signal_length = None  # initialize on first call
        self._scales  = None        # initialize on first call
        self._filters = None        # initialize on first call

    @abstractmethod
    def cwt(self, x):
        raise NotImplementedError

    def _build_filters(self):
        """
        Determines the optimal scale distribution (see. Torrence & Combo, Eq. 9-10),
        and then initializes the filter bank consisting of rescaled versions
        of the mother wavelet. Also includes normalization. Code is based on:
        https://github.com/aaren/wavelets/blob/master/wavelets/transform.py#L88
        """
        self._scale_minimum = self.compute_minimum_scale()
        self._scales = self.compute_optimal_scales()

        self._filters = [None]*len(self.scales)
        for scale_idx, scale in enumerate(self._scales):
            # Number of points needed to capture wavelet
            M = 10 * scale / self.dt
            # Times to use, centred at zero
            t = np.arange((-M + 1) / 2., (M + 1) / 2.) * self.dt
            if len(t) % 2 == 0: t = t[0:-1]  # requires odd filter size
            # Sample wavelet and normalise
            norm = (self.dt / scale) ** .5
            self._filters[scale_idx] = norm * self.wavelet(t, scale)

    def compute_optimal_scales(self):
        """
        Determines the optimal scale distribution (see. Torrence & Combo, Eq. 9-10).
        :return: np.ndarray, collection of scales
        """
        if self.signal_length is None:
            raise ValueError('Please specify signal_length before computing optimal scales.')
        J = int((1 / self.dj) * np.log2(self.signal_length * self.dt / self._scale_minimum))
        scales = self._scale_minimum * 2 ** (self.dj * np.arange(0, J + 1))
        return scales

    def compute_minimum_scale(self):
        """
        Choose s0 so that the equivalent Fourier period is 2 * dt.
        See Torrence & Combo Sections 3f and 3h.
        :return: float, minimum scale level
        """
        dt = self.dt
        def func_to_solve(s):
            return self.fourier_period(s) - 2 * dt
        return scipy.optimize.fsolve(func_to_solve, 1)[0]

    def power(self, x):
        """
        Performs CWT and converts to a power spectrum (scalogram).
        See Torrence & Combo, Section 4d.
        :param x: np.ndarray, batch of input signals of shape [n_batch,signal_length]
        :return: np.ndarray, scalogram for each signal [n_batch,n_scales,signal_length]
        """
        if self.unbias:
            return (np.abs(self.cwt(x)).T ** 2 / self.scales).T
        else:
            return np.abs(self.cwt(x)) ** 2

    @property
    def dt(self):
        return self._dt

    @dt.setter
    def dt(self, value):
        # Needs to recompute scale distribution and filters
        self._dt = value
        self._build_filters()

    @property
    def signal_length(self):
        return self._signal_length

    @signal_length.setter
    def signal_length(self, value):
        # Needs to recompute scale distribution and filters
        self._signal_length = value
        self._build_filters()

    @property
    def wavelet(self):
        return self._wavelet

    @property
    def fourier_period(self):
        """ Return a function that calculates the equivalent Fourier. """
        return getattr(self.wavelet, 'fourier_period')

    @property
    def scale_from_period(self):
        """ Return a function that calculates the wavelet scale from the fourier period """
        return getattr(self.wavelet, 'scale_from_period')

    @property
    def fourier_periods(self):
        """ Return the equivalent Fourier periods for the scales used. """
        assert self._scales is not None, 'Wavelet scales are not initialized.'
        return self.fourier_period(self.scales)

    @property
    def fourier_frequencies(self):
        """ Return the equivalent frequencies. """
        return np.reciprocal(self.fourier_periods)

    @property
    def scales(self):
        return self._scales

    @property
    def dj(self):
        return self._dj

    @property
    def wavelet(self):
        return self._wavelet

    @property
    def unbias(self):
        return self._unbias

    @property
    def complex_wavelet(self):
        return np.iscomplexobj(self._filters[0])

    @property
    def output_dtype(self):
        return np.complex128 if self.complex_wavelet else np.float64

##########################################################################################

class WaveletTransform(WaveletTransformBase):

    def __init__(self, dt=1.0, dj=0.125, wavelet=Morlet(), unbias=False):
        """
        This is SciPy version of the CWT filter bank. Main work for this filter bank
        is performed by the convolution implementated in 'scipy.signal.convolve'

        :param dt: float, sample spacing
        :param dj: float, scale distribution parameter
        :param wavelet: wavelet object, see 'wavelets.py'
        :param unbias: boolean, whether to unbias the power spectrum
        """
        super(WaveletTransform,self).__init__(dt, dj, wavelet, unbias)

    def cwt(self, x):
        """
        Implements the continuous wavelet transform on a batch of signals. All signals
        in the batch must have the same length, otherwise manual zero padding has to be
        applied. On the first call, the signal length is used to determines the optimal
        scale distribution and uses this for initialization of the wavelet filter bank.
        If there is only one example in the batch the batch dimension is squeezed.

        :param x: np.ndarray, batch of signals of shape [n_batch,signal_length]
        :return: np.ndarray, CWT for each signal in the batch [n_batch,n_scales,signal_length]
        """

        # Append batch dimension
        if x.ndim == 1:
            x = x[None,:]

        num_examples  = x.shape[0]
        signal_length = x.shape[-1]

        if signal_length != self.signal_length or not self._filters:
            # First call initializtion, or change in signal length. Note that calling
            # this also determines the optimal scales and initialized the filter bank.
            self.signal_length = signal_length

        # Wavelets can be complex so output is complex (np.float64 or np.complex128)
        cwt = np.zeros((num_examples, len(self.scales), x.shape[-1]), self.output_dtype)
        for example_idx in range(num_examples):
            cwt[example_idx] = self._compute_single(x[example_idx])

        # Squeeze batch dimension if single example
        if num_examples == 1:
            cwt = cwt.squeeze(0)
        return cwt

    def _compute_single(self, x):
        assert x.ndim == 1, 'Input signal must have single dimension.'
        output = np.zeros((len(self.scales), len(x)), self.output_dtype)
        for scale_idx, filt in enumerate(self._filters):
            output[scale_idx,:] = scipy.signal.convolve(x, filt, mode='same')
        return output

##########################################################################################

class WaveletTransformTorch(WaveletTransformBase):

    def __init__(self, dt=1.0, dj=0.125, wavelet=Morlet(), unbias=False, cuda=True):
        """
        This is PyTorch version of the CWT filter bank. Main work for this filter bank
        is performed by the convolution implementated in 'torch.nn.Conv1d'. Actual
        convolutions are performed by the helper class defined in 'network.py' which
        implements a 'torch.nn.module' that contains the convolution filters.

        :param dt: float, sample spacing
        :param dj: float, scale distribution parameter
        :param wavelet: wavelet object, see 'wavelets.py'
        :param unbias: boolean, whether to unbias the power spectrum
        :param cuda: boolean, whether to run convolutions on the GPU
        """
        super(WaveletTransformTorch, self).__init__(dt, dj, wavelet, unbias)
        self._cuda = cuda
        self._extractor = TorchFilterBank(self._filters, cuda)

    def cwt(self, x):
        """
        Implements the continuous wavelet transform on a batch of signals. All signals
        in the batch must have the same length, otherwise manual zero padding has to be
        applied. On the first call, the signal length is used to determines the optimal
        scale distribution and uses this for initialization of the wavelet filter bank.
        If there is only one example in the batch the batch dimension is squeezed.

        :param x: np.ndarray, batch of signals of shape [n_batch,signal_length]
        :return: np.ndarray, CWT for each signal in the batch [n_batch,n_scales,signal_length]
        """

        if x.ndim == 1:
            # Append batch_size and chn_in dimensions
            # [signal_length] => [n_batch,1,signal_length]
            x = x[None,None,:]
        elif x.ndim == 2:
            # Just append chn_in dimension
            # [n_batch,signal_length] => [n_batch,1,signal_length]
            x = x[:,None,:]

        num_examples  = x.shape[0]
        signal_length = x.shape[-1]

        if signal_length != self.signal_length or not self._filters:
            # First call initializtion, or change in signal length. Note that calling
            # this also determines the optimal scales and initialized the filter bank.
            self.signal_length = signal_length

        # Move to GPU and perform CWT computation
        x = torch.from_numpy(x).type(torch.FloatTensor)
        x.requires_grad_(requires_grad=False)

        if self._cuda: x = x.cuda()
        cwt = self._extractor(x)

        # Move back to CPU
        cwt = cwt.detach()
        if self._cuda:  cwt = cwt.cpu()
        cwt = cwt.numpy()

        if self.complex_wavelet:
            # Combine real and imag parts, returns object of shape
            # [n_batch,n_scales,signal_length] of type np.complex128
            cwt = (cwt[:,:,0,:] + cwt[:,:,1,:]*1j).astype(self.output_dtype)
        else:
            # Just squeeze the chn_out dimension (=1) to obtain an object of shape
            # [n_batch,n_scales,signal_length] of type np.float64
            cwt = np.squeeze(cwt, 2).astype(self.output_dtype)

        # Squeeze batch dimension if single example
        if num_examples == 1:
            cwt = cwt.squeeze(0)
        return cwt

    @property
    def dt(self):
        return self._dt

    @dt.setter
    def dt(self, value):
        super(WaveletTransformTorch, self.__class__).dt.fset(self, value)
        self._extractor.set_filters(self._filters)

    @property
    def signal_length(self):
        return self._signal_length

    @signal_length.setter
    def signal_length(self, value):
        super(WaveletTransformTorch, self.__class__).signal_length.fset(self, value)
        self._extractor.set_filters(self._filters)


from collections import Counter

import numpy as np


def _embed(x, order=3, delay=1):
    """
    Time-delay embedding.

    Parameters
    ----------
    x : 1d-array, shape (n_times)
        Time series
    order : int
        Embedding dimension (order)
    delay : int
        Delay.

    Returns
    -------
    embedded : ndarray, shape (n_times - (order - 1) * delay, order)
        Embedded time-series.

    Notes
    -----
    For deprecation in next major release. Use `time_delay_embedding` instead
    """
    return time_delay_embedding(x, order, delay)


def time_delay_embedding(time_series, embedding_dimension, delay=1):
    """
    Time-delayed embedding.

    Parameters
    ----------
    time_series : np.ndarray
        The input time series, shape (n_times)
    embedding_dimension : int
        The embedding dimension (order).
    delay : int
        The delay between embedded points.

    Returns
    -------
    embedded : ndarray
        The embedded time series with shape (n_times - (order - 1) * delay, order).
    """
    series_length = len(time_series)
    embedded_series = np.empty((embedding_dimension, series_length - (embedding_dimension - 1) * delay))
    for i in range(embedding_dimension):
        embedded_series[i] = time_series[i * delay : i * delay + embedded_series.shape[1]]
    return embedded_series.T


def util_pattern_space(time_series, lag, dim):
    """
    Create a set of sequences with given lag and dimension

    Parameters
    ----------
    time_series : np.ndarray
        Vector or string of the sample data
    lag : int
        Lag between the beginning of sequences
    dim : int
        Dimension (number of patterns)

    Returns
    -------
    pattern_space: np.ndarray
        2D array of vectors

    Raises
    ------
    ValueError: If the lag is less than 1 or the result matrix exceeds the size limit.
    """
    n = len(time_series)

    if lag < 1:
        raise ValueError("Lag should be greater than or equal to 1.")

    if lag * dim > n:
        raise ValueError("Result matrix size limit exceeded. Adjust the lag or dim value.")

    pattern_space = np.zeros((n - lag * (dim - 1), dim))
    for i in range(dim):
        pattern_space[:, i] = time_series[i * lag : i * lag + n - lag * (dim - 1)]

    return pattern_space


def util_granulate_time_series(time_series, scale):
    """
    Extract coarse-grained time series

    Parameters
    ----------
    time_series : np.ndarray
        Time series
    scale : int
        Scale factor

    Returns
    -------
    cts : np.ndarray
        Array of coarse-grained time series with given scale factor
    """
    if not isinstance(time_series, np.ndarray):
        time_series = np.array(time_series)

    n = time_series.shape[0]
    b = n // scale
    cts = np.mean(time_series[: b * scale].reshape(b, scale), axis=1)
    return cts


def shannon_entropy(time_series):
    """
    Return the Shannon Entropy of the sample data.

    Parameters
    ----------
    time_series: np.ndarray | list[str]
        Vector or string of the sample data

    Returns
    -------
    ent: float
        The Shannon Entropy as float value
    """
    if isinstance(time_series, str):
        # Calculate frequency counts
        counter = Counter(time_series)
        total_count = len(time_series)

        # Calculate frequencies and Shannon entropy
        ent = 0.0
        for count in counter.values():
            freq = count / total_count
            ent += freq * np.log2(freq)

        ent = -ent
        return ent

    # Calculate frequency counts
    _, counts = np.unique(time_series, return_counts=True)
    total_count = len(time_series)

    # Calculate frequencies and Shannon entropy
    frequencies = counts / total_count
    ent = -np.sum(frequencies * np.log2(frequencies))

    return ent


def sample_entropy(time_series, sample_length, tolerance=None):
    """
    Calculates the sample entropy of degree m of a time_series.

    This method uses Chebyshev norm.
    It is quite fast for random data, but can be slower is there is
    structure in the input time series.

    Parameters
    ----------
    time_series : np.ndarray
        Time series, 1-d vector
    sample_length : int
        length of longest template vector
    tolerance : float
        tolerance (defaults to 0.1 * std(time_series)))

    Returns
    -------
    sampen: np.ndarray
        Array of Sample Entropies SE.
        SE[k] is ratio `#templates of length k+1` / `#templates of length k`
        where `#templates of length 0` = n*(n - 1) / 2, by definition

    Notes
    -----
    The parameter 'sample_length' is equal to m + 1 in Ref[1].

    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/Sample_Entropy
    .. [2] http://physionet.incor.usp.br/physiotools/sampen/
    .. [3] Madalena Costa, Ary Goldberger, CK Peng. Multiscale entropy analysis of biological signals
    """
    if not isinstance(time_series, np.ndarray):
        time_series = np.array(time_series)

    if tolerance is None:
        tolerance = 0.1 * np.std(time_series)

    # The code below follows the sample length convention of Ref [1] so:
    sample_length = sample_length - 1
    n = len(time_series)

    # N_temp is a vector that holds the number of matches
    # N_temp[k] holds matches templates of length k
    N_temp = np.zeros(sample_length + 2)

    # Templates of length 0 matches by definition:
    N_temp[0] = n * (n - 1) / 2

    for i in range(n - sample_length - 1):
        template = time_series[i : (i + sample_length + 1)]  # We have `sample_length+1` elements in the template
        rem_time_series = time_series[i + 1 :]

        search_list = np.arange(len(rem_time_series) - sample_length, dtype=np.int32)
        for length in range(1, len(template) + 1):
            hit_list = np.abs(rem_time_series[search_list] - template[length - 1]) < tolerance
            N_temp[length] += np.sum(hit_list)
            search_list = search_list[hit_list] + 1

    sampen = -np.log(N_temp[1:] / N_temp[:-1])
    return sampen


def multiscale_entropy(time_series, sample_length, tolerance=None, maxscale=None):
    """
    Calculate the Multiscale Entropy of the given time series considering
    different time-scales of the time series.

    Parameters
    ----------
    time_series : np.ndarray
        Input time series for analysis.
    sample_length : int
        Bandwidth or group of points
    tolerance : float
        Tolerance value (default is 0.1 times the standard deviation of the `time_series`)
    maxscale : int
        Maximum timescale (default is the length of the `time_series`)

    Returns
    -------
    mse : np.ndarray
        Array of Multiscale Entropies

    References
    ----------
    .. [1] http://en.pudn.com/downloads149/sourcecode/math/detail646216_en.html
            Can be viewed at https://web.archive.org/web/20170207221539/http://en.pudn.com/downloads149/sourcecode/math/detail646216_en.html
    """
    if tolerance is None:
        tolerance = 0.1 * np.std(time_series)

    if maxscale is None:
        maxscale = len(time_series)

    mse = np.zeros(maxscale)
    for i in range(maxscale):
        temp = util_granulate_time_series(time_series, i + 1)
        mse[i] = sample_entropy(temp, sample_length, tolerance)[-1]
    return mse


def permutation_entropy(time_series, order=3, delay=1, normalize=False):
    """
    Permutation Entropy.

    Parameters
    ----------
    time_series : list | np.array
        Time series
    order : int
        Order of permutation entropy
    delay : int
        Time delay
    normalize : bool
        If True, divide by log2(factorial(m)) to normalize the entropy
        between 0 and 1. Otherwise, return the permutation entropy in bit.

    Returns
    -------
    pe : float
        Permutation Entropy

    References
    ----------
    .. [1] Massimiliano Zanin et al. Permutation Entropy and Its Main
        Biomedical and Econophysics Applications: A Review.
        http://www.mdpi.com/1099-4300/14/8/1553/pdf

    .. [2] Christoph Bandt and Bernd Pompe. Permutation entropy — a natural
        complexity measure for time series.
        http://stubber.math-inf.uni-greifswald.de/pub/full/prep/2001/11.pdf

    Notes
    -----
    Last updated (Oct 2018) by Raphael Vallat (raphaelvallat9@gmail.com):
    - Major speed improvements
    - Use of base 2 instead of base e
    - Added normalization

    Examples
    --------
    1. Permutation entropy with order 2

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value between 0 and log2(factorial(order))
        >>> print(permutation_entropy(x, order=2))
            0.918

    2. Normalized permutation entropy with order 3

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value comprised between 0 and 1.
        >>> print(permutation_entropy(x, order=3, normalize=True))
            0.589
    """
    x = np.array(time_series)
    hashmult = np.power(order, np.arange(order))
    # Embed x and sort the order of permutations
    sorted_idx = time_delay_embedding(x, embedding_dimension=order, delay=delay).argsort(kind="quicksort")
    # Associate unique integer to each permutations
    hashval = (np.multiply(sorted_idx, hashmult)).sum(1)
    # Return the counts
    _, c = np.unique(hashval, return_counts=True)
    # Use np.true_divide for Python 2 compatibility
    p = np.true_divide(c, c.sum())
    pe = -np.multiply(p, np.log2(p)).sum()
    if normalize:
        pe /= np.log2(np.math.factorial(order))
    return pe


def multiscale_permutation_entropy(time_series, m, delay, scale):
    """
    Calculate the Multiscale Permutation Entropy

    Parameters
    ----------
    time_series : np.ndarray
        Time series for analysis
    m : int
        Order of permutation entropy
    delay : int
        Time delay
    scale : int
        Scale factor

    Returns
    -------
    mspe : np.ndarray
        Array of Multiscale Permutation Entropies

    References
    ----------
    .. [1] Francesco Carlo Morabito et al. Multivariate Multi-Scale Permutation Entropy for
            Complexity Analysis of Alzheimer`s Disease EEG. www.mdpi.com/1099-4300/14/7/1186
    .. [2] http://www.mathworks.com/matlabcentral/fileexchange/37288-multiscale-permutation-entropy-mpe/content/MPerm.m
    """
    mspe = np.empty(scale)
    for i in range(scale):
        coarse_time_series = util_granulate_time_series(time_series, i + 1)
        mspe[i] = permutation_entropy(coarse_time_series, order=m, delay=delay)
    return mspe


def weighted_permutation_entropy(time_series, order=2, delay=1, normalize=False):
    """
    Calculate the weighted permutation entropy. Weighted permutation
    entropy captures the information in the amplitude of a signal where
    standard permutation entropy only measures the information in the
    ordinal pattern, "motif".

    Parameters
    ----------
    time_series : list | np.ndarray
        Time series
    order : int
        Order of permutation entropy
    delay : int
        Time delay
    normalize : bool
        If True, divide by log2(factorial(m)) to normalize the entropy
        between 0 and 1. Otherwise, return the permutation entropy in bit.

    Returns
    -------
    wpe : float
        Weighted Permutation Entropy

    References
    ----------
    .. [1] Bilal Fadlallah, Badong Chen, Andreas Keil, and José Príncipe
           Phys. Rev. E 87, 022911 - Published 20 February 2013

    Notes
    -----
    - Updated in Jun 2023 by Nikolay Donets
    - Updated in March 2021 by Samuel Dotson (samgdotson@gmail.com)

    Examples
    --------
    1. Weighted permutation entropy with order 2

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value between 0 and log2(factorial(order))
        >>> print(permutation_entropy(x, order=2))
            0.912

    2. Normalized weighted permutation entropy with order 3

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value comprised between 0 and 1.
        >>> print(permutation_entropy(x, order=3, normalize=True))
            0.547
    """
    x = time_delay_embedding(time_series, embedding_dimension=order, delay=delay)

    weights = np.var(x, axis=1)
    sorted_idx = x.argsort(kind="quicksort", axis=1)

    motif_weights = {}
    for weight, indices in zip(weights, sorted_idx):  # noqa: B905
        motif = tuple(indices)
        if motif in motif_weights:
            motif_weights[motif] += weight
        else:
            motif_weights[motif] = weight

    pw = np.array(list(motif_weights.values()))
    pw /= weights.sum()

    b = np.log2(pw)
    wpe = -np.dot(pw, b)

    if normalize:
        wpe /= np.log2(np.math.factorial(order))

    return wpe


def composite_multiscale_entropy(time_series, sample_length, scale, tolerance=None):
    """
    Composite Multiscale Entropy of the given time series

    Parameters
    ----------
    time_series : np.ndarray
        Time series for analysis
    sample_length : int
        Number of sequential points of the time series
    scale : int
        Scale factor
    tolerance : float
        Tolerance (default = 0.1 * std(time_series))

    Returns
    -------
    cmse : np.ndarray
        Array of Composite Multiscale Entropies

    References
    ----------
    .. [1] Wu, Shuen-De, et al. "Time series analysis using
        composite multiscale entropy." Entropy 15.3 (2013): 1069-1084.
    """
    if tolerance is None:
        tolerance = 0.1 * np.std(time_series)

    cmse = np.zeros(scale)

    for i in range(scale):
        for j in range(i + 1):
            tmp = util_granulate_time_series(time_series[j:], i + 1)
            se = sample_entropy(tmp, sample_length, tolerance)[-1]
            cmse[i] += se / (i + 1)
    return cmse



import torch
import torch.nn as nn
import torch.nn.functional as F

def compute_size(input_size, conv_layer):
    kernel_size = conv_layer.kernel_size[0]
    dilation = conv_layer.dilation[0]
    stride = conv_layer.stride[0]
    padding = conv_layer.padding[0]
    new_size = int((input_size + 2*padding - dilation*(kernel_size - 1) - 1)//stride + 1)
    return new_size

class CNN1D(nn.Module):
    def __init__(self, input_size = 1024, in_channels = 1, out_channels = 64, kernel_size = 9, n_class = 7):
        torch.manual_seed(28)
        super().__init__()
        self.__type__ = 'cnn'
        self.__arch__ = 'cnn1d'
        self.__version__ = 'baseline'

        #Layer 1
        self.conv1 = nn.Conv1d(in_channels, out_channels = out_channels, kernel_size = kernel_size, dilation = 3, stride = 2)
        layer1_shape = compute_size(input_size, self.conv1)
        self.norm1 = nn.BatchNorm1d(out_channels)

        #Layer 2
        self.conv2 = nn.Conv1d(in_channels = out_channels, out_channels = 2*out_channels, kernel_size = kernel_size, dilation = 3, stride = 2)
        layer2_shape = compute_size(layer1_shape, self.conv2)
        self.norm2 = nn.BatchNorm1d(2*out_channels)

        #Layer 3
        self.conv3 = nn.Conv1d(in_channels = 2*out_channels, out_channels = out_channels, kernel_size = kernel_size, dilation = 3, stride = 2)
        layer3_shape = compute_size(layer2_shape, self.conv3)
        self.norm3 = nn.BatchNorm1d(out_channels)

        #Pool
        self.pool = nn.AvgPool1d(kernel_size=3)

        #Activation
        self.actv = nn.ReLU()

        #fully-connected
        flatten_shape = self.conv3.out_channels*(layer3_shape//3)

        self.fc = nn.Linear(flatten_shape, n_class)
        self.dropout = nn.Dropout(0.5)
        
        self.init_weights = False

    def forward(self,x):
        x = self.actv(self.conv1(x))
        x = self.norm1(x)

        x = self.actv(self.conv2(x))
        x = self.norm2(x)

        x = self.actv(self.conv3(x))
        x = self.norm3(x)

        x = self.pool(x)
        x = x.flatten(1)

        x = self.dropout(x)
        x = self.fc(x)
        
        x = F.log_softmax(x,-1)
        return x    



import numpy as np 
from scipy.stats import skew
import librosa 
from copy import deepcopy as dc

def calculate_statistics(list_values):
    n5 = np.nanpercentile(list_values, 5)
    n25 = np.nanpercentile(list_values, 25)
    n75 = np.nanpercentile(list_values, 75)
    n95 = np.nanpercentile(list_values, 95)
    median = np.nanpercentile(list_values, 50)
    mean = np.nanmean(list_values)
    std = np.nanstd(list_values)
    var = std**2
    rms = np.sqrt(np.nanmean(list_values**2))
    sk = skew(list_values)
    if np.isnan(sk):
        sk = 0
    zcr = librosa.feature.zero_crossing_rate(y = list_values,frame_length = len(list_values),center = False).item()
    en = shannon_entropy(list_values)
    perm_en = permutation_entropy(list_values)
    return [n5, n25, n75, n95, median, mean, std, var, rms, sk, zcr, en, perm_en]

def calculate_spectrum_statistics(list_values, n_fft = 1024, window_size = 1024, center = False):
    # fft  
    spectrum = np.abs(librosa.stft(list_values, n_fft = n_fft,center = center)).ravel()/window_size
    power_spectrum = spectrum**2

    spec_centroid = spectral_centroid(spectrum)
    spec_entropy = shannon_entropy(power_spectrum)
    spec_flatness = spectral_flatness(power_spectrum)
    return [spec_centroid, spec_entropy, spec_flatness]

def spectral_centroid(spectrum):
    return np.sum(spectrum*fft_freq) / np.sum(spectrum)

def spectral_flatness(spectrum):
    if 0. in spectrum:
        return 0
    else: 
        n = len(spectrum)
        return np.exp(np.mean(np.log(spectrum))/n)/(np.mean(spectrum)/n)
    
def shannon_entropy(array):
    max = array.max()
    min = array.min()
    if max == min:
        return np.log(2)
    else:
        normalized_array = (array - min)/(max - min)
        entropy = 0
        for i in range(len(array)):
            if normalized_array[i] != 0:
                entropy -= normalized_array[i]*np.log(normalized_array[i])
        return entropy

def time_delay_embedding(time_series, embedding_dimension, delay=1):
    """
    Time-delayed embedding.

    Parameters
    ----------
    time_series : np.ndarray
        The input time series, shape (n_times)
    embedding_dimension : int
        The embedding dimension (order).
    delay : int
        The delay between embedded points.

    Returns
    -------
    embedded : ndarray
        The embedded time series with shape (n_times - (order - 1) * delay, order).
    """
    series_length = len(time_series)
    embedded_series = np.empty((embedding_dimension, series_length - (embedding_dimension - 1) * delay))
    for i in range(embedding_dimension):
        embedded_series[i] = time_series[i * delay : i * delay + embedded_series.shape[1]]
    return embedded_series.T

def permutation_entropy(time_series, order=3, delay=1, normalize=False):
    """
    Permutation Entropy.

    Parameters
    ----------
    time_series : list | np.array
        Time series
    order : int
        Order of permutation entropy
    delay : int
        Time delay
    normalize : bool
        If True, divide by log2(factorial(m)) to normalize the entropy
        between 0 and 1. Otherwise, return the permutation entropy in bit.

    Returns
    -------
    pe : float
        Permutation Entropy

    References
    ----------
    .. [1] Massimiliano Zanin et al. Permutation Entropy and Its Main
        Biomedical and Econophysics Applications: A Review.
        http://www.mdpi.com/1099-4300/14/8/1553/pdf

    .. [2] Christoph Bandt and Bernd Pompe. Permutation entropy — a natural
        complexity measure for time series.
        http://stubber.math-inf.uni-greifswald.de/pub/full/prep/2001/11.pdf

    Notes
    -----
    Last updated (Oct 2018) by Raphael Vallat (raphaelvallat9@gmail.com):
    - Major speed improvements
    - Use of base 2 instead of base e
    - Added normalization

    Examples
    --------
    1. Permutation entropy with order 2

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value between 0 and log2(factorial(order))
        >>> print(permutation_entropy(x, order=2))
            0.918

    2. Normalized permutation entropy with order 3

        >>> x = [4, 7, 9, 10, 6, 11, 3]
        >>> # Return a value comprised between 0 and 1.
        >>> print(permutation_entropy(x, order=3, normalize=True))
            0.589
    """
    x = np.array(time_series)
    hashmult = np.power(order, np.arange(order))
    # Embed x and sort the order of permutations
    sorted_idx = time_delay_embedding(x, embedding_dimension=order, delay=delay).argsort(kind="quicksort")
    # Associate unique integer to each permutations
    hashval = (np.multiply(sorted_idx, hashmult)).sum(1)
    # Return the counts
    _, c = np.unique(hashval, return_counts=True)
    # Use np.true_divide for Python 2 compatibility
    p = np.true_divide(c, c.sum())
    pe = -np.multiply(p, np.log2(p)).sum()
    if normalize:
        pe /= np.log2(np.math.factorial(order))
    return pe

def GASF_GADF(arr):
    array = arr.reshape(-1,1)
    max = array.max()
    min = array.min()
    if max > min:
        normed_array = (2*array-max-min)/(max - min)
    else:
        normed_array = array/len(array)
    sin = np.sqrt(np.ones((len(array),1))-normed_array**2)
    coscos = normed_array@normed_array.T
    sinsin = sin@sin.T
    sincos = sin@array.T
    cossin = array@sin.T
    return np.array(coscos - sinsin, sincos - cossin)